/***************************************************************************
 * Artekit STM32 DIY lightsaber audio board
 * Copyright (C) 2013 Artekit Italy
 * Written by Ivan Meleca
 * http://www.artekit.eu
 
### flash.c
 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#include "flash.h"
#include "spi.h"
#include "dma.h"
#include "sys.h"

/* Reads the serial flash data register */
uint8_t flashGetStatusRegister(void)
{
	uint8_t res = 0;
	
	FLASH_CS_ON();
	spiWriteByte(0x05);
	res = spiReadByte();
	FLASH_CS_OFF();
	
	return res;
}

/* Checks if the serial flash is busy in a write operation */
uint8_t flashIsBusy(void)
{
	return ( flashGetStatusRegister() & 0x01 );
}

/* Enables write operations */
void flashWriteEnable(void)
{
	FLASH_CS_ON();
	spiWriteByte(0x06);
	FLASH_CS_OFF();
}

/* Disables write operations */
void flashWriteDisable(void)
{
	FLASH_CS_ON();
	spiWriteByte(0x04);
	FLASH_CS_OFF();
}

/* Unprotects all blocks */
void flashUnprotect(void)
{
	flashWriteEnable();
	
	FLASH_CS_ON();
	spiWriteByte(0x01);
	spiWriteByte(0x00);
	FLASH_CS_OFF();
}

/* Erases the entire serial flash */
void flashChipErase(void)
{
	while (flashIsBusy());
	
	flashUnprotect();
	
	flashWriteEnable();
	
	FLASH_CS_ON();
	spiWriteByte(0x60);
	FLASH_CS_OFF();
	
	while (flashIsBusy());
}

/* Reads 'qty' bytes from the 'addr' address into 'buffer' */
void flashRead(uint32_t addr, uint8_t* buffer, uint32_t qty)
{
	uint8_t dummy = 0;
	
	FLASH_CS_ON();
	spiWriteByte(0x0b);
	spiWriteByte( (uint8_t) ((addr >> 16) & 0xFF));
	spiWriteByte( (uint8_t) ((addr >> 8) & 0xFF));
	spiWriteByte( (uint8_t) (addr & 0xFF));
	spiWriteByte(dummy);
	spiRead(buffer, qty);
	FLASH_CS_OFF();
}

/* Queues a DMA request. Called from TIM1 when it's time to read an audio chunk */
uint8_t flashDMARequest(uint32_t addr, uint8_t* buffer, uint32_t qty, uint8_t* signal)
{
	DMA_REQUEST* req = 0;

	req = GetFreeReq();
	
	if (req == 0)
	{
		/* something is wrong */
		return 0;
	}

	req->active = 1;
	req->buffer = buffer;
	req->qty = qty;
	req->signal_ptr = signal;
	req->cmd[0] = 0x0B;
	req->cmd[1] = (uint8_t) ((addr >> 16) & 0xFF);
	req->cmd[2] = (uint8_t) ((addr >> 8) & 0xFF);
	req->cmd[3] = (uint8_t) (addr & 0xFF);
	req->cmd[4] = 0;
	
	ProcessRequest(req);
	
	return 1;
}

/* Writes 'qty' bytes from 'buffer' into the 'addr' address */
void flashWrite(uint32_t addr, uint8_t* buffer, uint32_t qty)
{
	uint32_t written = 0;
	uint8_t aai = 0;
	
	if (flashGetStatusRegister() & 0x40)
	{
		flashWriteDisable();
	}
	
	flashWriteEnable();
	
	while (written < qty)
	{
		while(flashIsBusy());
		
		if (qty - written >= 2)
		{
			FLASH_CS_ON();
			
			spiWriteByte(0xAD);
			
			if (aai == 0)
			{
				spiWriteByte( (uint8_t) ((addr >> 16) & 0xFF));
				spiWriteByte( (uint8_t) ((addr >> 8) & 0xFF));
				spiWriteByte( (uint8_t) (addr & 0xFF));
				aai = 1;
			}
			
			spiWriteByte(buffer[written++]);
			spiWriteByte(buffer[written++]);
			
			FLASH_CS_OFF();
			addr += 2;
		} else {
			
			if (aai)
			{
				flashWriteDisable();
				flashWriteEnable();
				aai = 0;
			}
			
			FLASH_CS_ON();
			
			spiWriteByte(0x02);
			spiWriteByte( (uint8_t) ((addr >> 16) & 0xFF));
			spiWriteByte( (uint8_t) ((addr >> 8) & 0xFF));
			spiWriteByte( (uint8_t) (addr & 0xFF));
			spiWriteByte(buffer[written++]);
			
			FLASH_CS_OFF();
		}
	}
	
	while(flashIsBusy());
	flashWriteDisable();
}

