/***************************************************************************
 * Artekit STM32 DIY lightsaber audio board
 * Copyright (C) 2013 Artekit Italy
 * Written by Ivan Meleca
 * http://www.artekit.eu
 
### main.c
 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#include "stm32f10x.h"
#include "sys.h"
#include "flash.h"
#include "sys.h"
#include "config.h"
#include "uart.h"
#include "spi.h"
#include "mma8452.h"
#include "dma.h"
#include <string.h>
#include <stdlib.h>

/*
 * What this code does:
 * Simulates lightsaber sound effects using very low cost components.
 *
 * How this code works (in a very quick detail, extracted from the blog):
 * There is the STM32, a serial flash, an accelerometer, a DAC, an amplifier and two buttons.
 * The STM32 configures the accelerometer to give indications of pulses and transients on two
 * interrupt pins of the STM32. Maths are done inside the accelerometer (by freescale), so the
 * I2C channel is free for whatever you want, after the startup.
 * 
 * There are some waveforms stored in the serial flash (16-bit PCM, 22050 sample rate).
 * These waveform are divided in two groups. The HUM, and the ALT: the HUM is the background sound
 * that is always on. The ALT is a sporadic sound that is mixed with the HUM sound. These ALT sounds
 * are divided in the following way: the ON sound, the OFF sound, the HIT sound (a laser from a
 * pistol hits the blade), the STRIKE sound (two saber blades that meet) and the SWING (sort of
 * a doppler simulation when the saber is moved quickly).
 * 
 * The STM32 will read the waveforms from the serial flash (through SPI) and send them to the DAC
 * (using a GPIO-toggled SPI, on TIM2).
 *
 * The complex part of this implementation is the reading of the waveforms. We use a double buffer
 * method: while we play a chunk of samples, in the meantime, we read (at 36Mhz from the serial
 * flash) the next chunk of samples. When the first chunk of data is done playing, we quickly switch
 * buffers and start playing the next chunk. The cycle repeats. Since we are playing 22050 samples
 * per second, the chunk size has to be long enough to keep playing while we read another chunk.
 * To add complexity, the said method is done for two kinds of sounds (ALT and HUM), that have to be
 * loaded and mixed in real time. So instead of two switching-buffers, we have four. This sums to
 * the time the chunk has to be playing while we grab (possibly) another two chunks (one for HUM,
 * one for ALT when it's playing).
 *
 * So we have TIM1 running at aprox. 22050 Hz. And TIM2 sending the last read (and mixed) sample
 * to the DAC. Since we only have 20KB of RAM and we don't want to be busy all the time inside a timer
 * interrupt, we use SPI with DMA. There is a "DMA request queue" mechanism defined in dma.c that 
 * appends DMA requests of sound chunks and execute them one after another, and flags the chucks as
 * "ready" when it's done reading from the serial flash. The serial flash is so fast that I believe
 * we are wasting more time setting the DMA stuff than actually reading.
 *
 * Buttons are simple: one for turning the thing on and off, and one for the HIT sound since we didn't
 * find a simpler way to simulate a laser pistol shooting at a blade we don't have :)
 *
 * Hardware used:
 * - STM32F103
 * - MMA8452Q Accelerometer from Freescale
 * - SST25VF080B serial flash from SST (Microchip)
 * - MCP4921 DAC from Microchip
 * - NJM2070D amplifier from JRC
 *
 */

/* Codes for buttons */
#define BUTTON_RELEASED		0
#define BUTTON_PRESSED		1

/* These are the bitbanding values for the CLK, DATA and CS of our simulated SPI to talk to the DAC
 * (since the real SPI is busy fetching waveforms from the serial flash).
 */
#define DAC_CLK     		(*( (volatile unsigned long *) 0x4221018C))
#define DAC_DATA    		(*( (volatile unsigned long *) 0x42210184))
#define DAC_CS 				(*( (volatile unsigned long *) 0x42210188))

/* Buffer-related */
/* Sample size = 2, signed 16 bit PCM audio format */
#define SND_BUF_SAMPLE_SIZE		2
#define SND_BUF_MAX_SAMPLES		1500

/* Precalculated quantity of bytes that can fit a single buffer */
#define SND_ALT_READ_BYTES		3000

/* For the HUM sound we actually read the half of an ALT buffer size:
 * Since the HUM is a played continuously, we need to make equal readings from the serial flash
 * (in matters of quantity of bytes). For example, if the HUM file is 10000 bytes length, we can't
 * read by chunks of 900 bytes, because the last reading will be of 100 bytes. And 100 bytes dont't
 * gives us time to read the 900 bytes of the chunk that has to be played next. So if the buffer
 * size is 900, we will read chunks of 450. While reading the last chunk we'll check if we can read
 * a little more than 450 bytes (550 bytes for the example above).
 * All this is necessary because we don't know the length of the HUM sound the user may use. Not in
 * a way we can predict static buffers.
 * For the ALT sound this is not needed since we don't care if the last chunk is 1 byte length. The
 * ALT sound is played once.
 */ 
#define SND_HUM_READ_SAMPLES	750
#define SND_HUM_READ_BYTES		1500

/* There are the indexes of the waveforms inside the serial flash */
extern SND_DESC snd_desc[SND_MAX_DESC];
extern SND_HEADER snd_header;

/* Sample reading and mixing */
static int16_t last_sample = 0;
static volatile uint16_t dac_sample = 0;
static volatile uint8_t dac_bits = 0;

/* We select a random SWING, HIT, and STRIKE, and save the index of last played to avoid playing
 * the same sound every time (it will happen).
 */
uint8_t last_swing = 0;
uint8_t last_hit = 0;
uint8_t last_strike = 0;

/* Struct to keep trace of button debouncing and state */
typedef struct button_struct
{
	uint32_t ticks;
	uint32_t press_time;
	uint32_t release_time;
	uint8_t state;
	GPIO_TypeDef* port;
	uint16_t pin;
} BUTTON;

/* Struct that holds a chunk of waveforms */
typedef struct snd_buf_struct
{
	uint8_t loaded;						/* We give a pointer to this variable to the DMA request queue.
										 * Upon reading completion, the DMA will set this variable to 1
										 * indicating that the chunk is loaded.
										 */

	uint16_t buf[SND_BUF_MAX_SAMPLES];	/* Waveform buffer */
	uint16_t sample_offset;				/* The offset of the sample we are playing */
	uint16_t samples_loaded;			/* The quantity of samples loaded */
} SND_BUF;

/* Struct that holds the HUM audio information */
typedef struct hum_playback_struct
{
	SND_BUF buffers[2];			/* Two buffers */
	SND_BUF* playing;			/* Pointer to the buffer we're playing */
	SND_BUF* loading;			/* Pointer to the buffer we're loading */
	volatile uint8_t do_play;	/* 1 when we can start playing */
	uint32_t total_samples;		/* Total samples of the waveform file */
	uint32_t samples_left;		/* Samples to play */
	uint32_t start_addr;		/* Starting address on the serial flash */
	uint32_t next_addr;			/* Next address to load from the serial flash */
	
} HUM_PLAYBACK;

/* Struct that holds the ALT audio information */
typedef struct alt_playback_struct
{
	SND_BUF buffers[2];			/* Two buffers */
	SND_BUF* playing;			/* Pointer to the buffer we're playing */
	SND_BUF* loading;			/* Pointer to the buffer we're loading */

	volatile uint8_t do_play;	/* 1 when we can start playing */
	uint8_t stop_signal;		/* 1 when we need to abort */
	uint32_t samples_left;		/* Total samples to be played */
	uint32_t next_addr;			/* Next flash address to load */
	uint8_t type;				/* Alt type playing */

} ALT_PLAYBACK;

/* HUM audio struct */
static HUM_PLAYBACK hum_playback;

/* ALT audio struct */
static ALT_PLAYBACK alt_playback;

/* For debug only */
static uint32_t timer_missed = 0;

/* Button ON/OFF connected on PA12 */
static BUTTON button_on =
{
	0,				/* tick */
	0,				/* time */
	0,				/* action */
	BUTTON_RELEASED,/* state */
	GPIOA,			/* port */
	12,				/* pin */
};

/* Extra buttong connected to PB5 */
static BUTTON button_extra =
{
	0,				/* tick */
	0,				/* time */
	0,				/* action */
	BUTTON_RELEASED,/* state */
	GPIOB,			/* port */
	5,				/* pin */
};

/* This function switches and loads the chunks for the HUM audio */
static void SwitchHumBuffers(void)
{
	/* Do the switching */
	SND_BUF* prev = hum_playback.playing;
	hum_playback.playing = hum_playback.loading;
	hum_playback.loading = prev;
	
	/* Reset the variables of the chunk we're loading */
	hum_playback.loading->loaded = 0;
	hum_playback.loading->sample_offset = 0;
	
	/* Check if we have to load the last chunk */
	if (hum_playback.samples_left < SND_BUF_MAX_SAMPLES)
	{
		/* Read the remaining samples, even if they're longer than SND_HUM_READ_SAMPLES */
		flashDMARequest(hum_playback.next_addr,
						(uint8_t*) hum_playback.loading->buf,
						hum_playback.samples_left * SND_BUF_SAMPLE_SIZE,
						&hum_playback.loading->loaded);
		
		/* Since it's the last chunk, point everything back to the beginning of the waveform */
		hum_playback.loading->samples_loaded = hum_playback.samples_left;
		hum_playback.next_addr = hum_playback.start_addr;
		hum_playback.samples_left = hum_playback.total_samples;
	
	} else {
		
		/* Read SND_HUM_READ_SAMPLES samples */
		flashDMARequest(hum_playback.next_addr,
						(uint8_t*) hum_playback.loading->buf,
						SND_HUM_READ_BYTES,
						&hum_playback.loading->loaded);
		
		/* Update pointer and counters */
		hum_playback.next_addr += SND_HUM_READ_BYTES;
		hum_playback.loading->samples_loaded = SND_HUM_READ_SAMPLES;
		hum_playback.samples_left -= SND_HUM_READ_SAMPLES;
	}
}

/* This function switches and loads the chunks for the ALT audio */
static void SwitchAltBuffers(void)
{
	/* Do the switching */
	SND_BUF* prev = alt_playback.playing;
	alt_playback.playing = alt_playback.loading;
	alt_playback.loading = prev;
	
	/* Reset the variables of the chunk we're loading */
	alt_playback.loading->loaded = 0;
	alt_playback.loading->sample_offset = 0;
	
	/* Read always SND_ALT_READ_BYTES bytes, even if we're going EOF. We should play
	 * only "alt_playback.samples_left" samples.
	 */
	flashDMARequest(alt_playback.next_addr,
					(uint8_t*) alt_playback.loading->buf,
					SND_ALT_READ_BYTES,
					&alt_playback.loading->loaded);
	
	alt_playback.next_addr += SND_ALT_READ_BYTES;
	alt_playback.loading->samples_loaded = SND_BUF_MAX_SAMPLES;
}

/* This function mixes the HUM and ALT samples */
static int16_t MixSamples(int16_t hum, int16_t alt)
{
	int32_t sum = 0;
	int16_t ret;
	
	/* Uncomment this if you want the (A + B) / 2 method */
	/*
	sum = (hum + alt) / 2;
	ret = sum;
	*/
	
	sum = hum + alt;
	if (sum < (int16_t) 0x8000)
	{
		ret = 0x8000;
	} else if (sum > (int16_t) 0x7FFF)
	{
		ret = 0x7FFF;
	} else {
		ret = (int16_t) sum;
	}
	
	return ret;
}

/* TIM1 IRQ handler. Does the audio loading and buffer switching */
void TIM1_UP_IRQHandler(void)
{
	int16_t alt_sample = 0;
	int16_t hum_sample = 0;
	
	/* Reset IRQ flag */
	TIM1->SR = 0xFFFE;
	
	/* Has to play the HUM sound? */
	if (hum_playback.do_play)
	{
		/* Do we have at least a chunk to play */
		if (hum_playback.playing->loaded)
		{
			/* Have we finished the samples to play? */
			if (hum_playback.playing->sample_offset == hum_playback.playing->samples_loaded)
			{
				/* Yes, switch buffers and load another chunk. */
				SwitchHumBuffers();
			}
			
			/* Store the sample to play into the hum_sample variable */
			hum_sample = hum_playback.playing->buf[hum_playback.playing->sample_offset++];
		} else {
			/* We are missing samples!!! */
			timer_missed++;
		}
	}

	/* Has to play the ALT sound? */
	if (alt_playback.do_play)
	{
		/* Do we have at least a chunk to play */
		if (alt_playback.playing->loaded)
		{
			/* Have we finished the samples to play?
			 * Is the main program stopping the ALT playback?
			 */
			if (!alt_playback.samples_left || alt_playback.stop_signal)
			{
				/* Stop playing */
				alt_playback.do_play = 0;
			} else {
				/* Are we done with this audio chunk? */
				if (alt_playback.playing->sample_offset == SND_BUF_MAX_SAMPLES)
				{
					/* Yes, switch buffers and load another chunk */
					SwitchAltBuffers();
				}
					
				/* Decrement the samples we have left to play */
				alt_playback.samples_left--;

				/* Store the sample to play into the alt_sample variable */
				alt_sample = alt_playback.playing->buf[alt_playback.playing->sample_offset++];
			}
		} else {
			/* We are missing samples!!! */
			timer_missed++;
		}
	}

	/* Mix the ALT and HUM samples into the last_sample variable */
	last_sample = MixSamples(hum_sample, alt_sample);
	
	/* Prepare the dac_sample variable. This is what we'll be sending to the DAC */
	
	/* Sign and 12-bit calculation */
	dac_sample = (uint16_t) last_sample + (uint16_t) 0x8000;
	dac_sample = (dac_sample >> 4) & (uint16_t) 0xFFF;
	
	/* Set the DAC enable flag */
	dac_sample |= 0x3000;
	
	/* Set the quantity of bits to send */
	dac_bits = 16;
	
	/* Pull down chip select */
	DAC_CS = 0;
	
	/* Enable TIM2 */
	TIM2->CR1 |= TIM_CR1_CEN;
}

/* Here we brute-force a SPI by toggling GPIOs */
void TIM2_IRQHandler(void)
{
	/* Reset IRQ flag */
	TIM2->SR = 0xFFFE;
	
	/* CLK low */
	DAC_CLK = 0;
	
	/* Put DATA pin */
	DAC_DATA = (dac_sample >> --dac_bits) & 0x01;
	
	/* CLK high */
	DAC_CLK = 1;
	
	/* Done? */
	if (dac_bits == 0)
	{
		/* Disable timer and put the pins in a known state */
		TIM2->CR1 &= ~TIM_CR1_CEN;
		DAC_CS = 1;
		DAC_CLK = 0;
		DAC_DATA = 0;
		dac_sample = 0;
	}
}

/* This function is called when the HUM sound has to be played */
static void PlayHum(void)
{
	uint8_t i;
	SND_DESC* hum = NULL;

	/* Already playing? */
	if (hum_playback.do_play) return;
	
	/* Find the HUM sound */
	for (i = 0; i < snd_header.cnt; i++)
	{
		if (snd_desc[i].type == SND_TYPE_HUM)
		{
			hum = &snd_desc[i];
		}
	}
	
	if (!hum)
	{
		/* Something is wrong */
		return;
	}
	
	/* Fill the playback structure with the information we need */	
	hum_playback.start_addr 		= hum_playback.next_addr = hum->offs;	
	hum_playback.playing->loaded 	= 0;
	hum_playback.loading->loaded	= 0;
	hum_playback.total_samples 		= hum->len / SND_BUF_SAMPLE_SIZE;
	hum_playback.samples_left 		= hum_playback.total_samples;
	hum_playback.playing->samples_loaded = SND_HUM_READ_SAMPLES;
	hum_playback.loading->samples_loaded = SND_HUM_READ_SAMPLES;
	hum_playback.playing->sample_offset = 0;
	hum_playback.loading->sample_offset = 0;
	
	/* Load the first chunk */
	flashDMARequest(hum_playback.next_addr,
					(uint8_t*) hum_playback.playing->buf,
					SND_HUM_READ_BYTES,
					&hum_playback.playing->loaded);
	
	hum_playback.next_addr += SND_HUM_READ_BYTES;
	hum_playback.samples_left -= SND_HUM_READ_SAMPLES;
	
	/* Load the second chunk */
	flashDMARequest(hum_playback.next_addr,
					(uint8_t*) hum_playback.loading->buf,
					SND_HUM_READ_BYTES,
					&hum_playback.loading->loaded);

	hum_playback.next_addr += SND_HUM_READ_BYTES;
	hum_playback.samples_left -= SND_HUM_READ_SAMPLES;

	/* Give some time to the DMA to finish */
	sysDelayMs(100);
	
	/* Start playing. The TIM1 timer checks this variable to know if it has to play the HUM sound */
	hum_playback.do_play = 1;
}

/* Stops playing the HUM sound */
static void StopHum(void)
{
	hum_playback.do_play = 0;
}

/* This function is called when an ALT sound has to be played */
static void StartAlt(uint8_t num)
{
	/* Already playing? */
	if (alt_playback.do_play)
	{
		alt_playback.stop_signal = 1;
		/* Wait for the timer to finish */
		while (alt_playback.do_play);
		
		/* wait for dma requests */
		while (!alt_playback.playing->loaded && !alt_playback.loading->loaded);
	}
	
	/* Fill the playback structure with the information we need */
	alt_playback.samples_left = snd_desc[num].len / SND_BUF_SAMPLE_SIZE;
	alt_playback.next_addr = snd_desc[num].offs;
	alt_playback.playing->samples_loaded = SND_BUF_MAX_SAMPLES;
	alt_playback.loading->samples_loaded = SND_BUF_MAX_SAMPLES;
	alt_playback.playing->sample_offset = 0;
	alt_playback.loading->sample_offset = 0;
	alt_playback.stop_signal = 0;
	alt_playback.type = snd_desc[num].type;
	
	/* We assume there are at least two chunks to play */
	
	/* Load the first chunk */
	flashDMARequest(alt_playback.next_addr,
					(uint8_t*) alt_playback.playing->buf,
					SND_ALT_READ_BYTES,
					&alt_playback.playing->loaded);
	
	alt_playback.next_addr += SND_ALT_READ_BYTES;
	
	/* Load the second chukn chunk */
	flashDMARequest(alt_playback.next_addr,
					(uint8_t*) alt_playback.loading->buf,
					SND_ALT_READ_BYTES,
					&alt_playback.loading->loaded);

	alt_playback.next_addr += SND_ALT_READ_BYTES;
	
	/* Start playing. The TIM1 timer checks this variable to know if it has to play the ALT sound */
	alt_playback.do_play = 1;
}

/* This function is called to pick a pseudo-random ALT sound to play.
 * It keeps track of the last played ALT for every type, so the same ALT sound isn't played twice
 */
static void PlayRandomAlt(uint8_t type)
{
	uint8_t val = 0;
	uint8_t i;
	uint8_t min = 0;
	uint8_t max = 0;
	uint8_t* last = 0;
	
	switch (type)
	{
		case SND_TYPE_HIT:		last = &last_hit; 		break;
		case SND_TYPE_STRIKE:	last = &last_strike;	break;
		case SND_TYPE_SWING:	last = &last_swing;		break;
		default: return;
	}

	for (i = 0; i < snd_header.cnt; i++)
	{
		if (snd_desc[i].type == type)
		{
			if (i < min || !min)
			{
				min = i;
			}
			
			if (i > max || !max)
			{
				max = i;
			}
		}
	}
	
	srand(sysGetTickCount());
	val = (rand() % (max-min))+min;
	
	if (*last == val || *last == val - 1)
	{
		if (val == max)
		{
			val = min;
		} else {
			if (max - min)
			{
				val++;
			}
		}
	}
	
	*last = val;
	
	StartAlt(val);
	return;		
}

/* Plays the ON sound */
static void PlayOn(void)
{
	uint8_t i;

	/* Find the ON sound */
	for (i = 0; i < snd_header.cnt; i++)
	{
		if (snd_desc[i].type == SND_TYPE_ON)
		{
			StartAlt(i);
			return;
		}
	}
}

/* Plays the OFF sound */
static void PlayOff(void)
{
	uint8_t i;

	/* Find the OFF sound */
	for (i = 0; i < snd_header.cnt; i++)
	{
		if (snd_desc[i].type == SND_TYPE_OFF)
		{
			StartAlt(i);
			return;
		}
	}
}

/* Debounces the button press */
static uint8_t GetButtonTime(BUTTON* button, uint8_t action, uint32_t time)
{
	uint8_t cur_state = !((button->port->IDR >> button->pin) & 0x01);
	
	if (action == BUTTON_RELEASED)
	{
		/* wait to be pressed */
		if (cur_state)
		{
			button->release_time = sysGetTickCount();
		} else {
			if (button->release_time != 0)
			{
				if (sysGetTickCount() - button->release_time > time)
				{
					button->release_time = 0;
					return 1;
				}
			}
		}
	} else if (action == BUTTON_PRESSED)
	{
		/* wait to be released */
		if (!cur_state)
		{
			button->press_time = sysGetTickCount();
		} else {
			if (button->press_time != 0)
			{
				if (sysGetTickCount() - button->press_time > time)
				{
					button->press_time = 0;
					return 1;
				}
			}
		}
	}
	
	return 0;
}

int main(void)
{
	uint8_t state = 0;
	
	/* Init clocks, GPIO and SysTick */
	sysRCCConfiguration();
	sysConfigGPIO();
	sysInitSystickTimer();
	
	/* Init SPI and UART */
	spiInit();
	uartInit();
	
	/* Check if we have a valid configuration.
	 * To know about serial flash configuration check the config.c file
	 */
	if (!configLoad()) goto LOAD;
	
	/* Start and configure accelerometer */
	accInit();
	
	/* Configure TIM1 and TIM2 */
	sysInitTimer();
	
	/* Point the playback buffer pointers to their repective initial buffers */
	alt_playback.playing = &alt_playback.buffers[0];
	alt_playback.loading = &alt_playback.buffers[1];
			
	hum_playback.playing = &hum_playback.buffers[0];
	hum_playback.loading = &hum_playback.buffers[1];
	
	/* Check incoming characters from the serial port.
	 * If we have a char means we enter configuration mode.
	 */
	while (!configPoll())
	{
		/* Two states. On and Off */
		switch (state)
		{
			case 0:
				/* state = 0 --> OFF */
			
				/* Detect if the ON/OFF button is pressed for more than 100 ms */
				if (GetButtonTime(&button_on, BUTTON_PRESSED, 100))
				{
					/* Play the ON sound */
					PlayOn();
					
					/* Play the HUM sound */
					PlayHum();
					
					/* State = ON */
					state++;
					
					/* Reset previous accelerometer indications that may have happened while
					 * we were in the OFF state
					 */
					accResetStatusFlags();
				}	
				break;
				
			case 1:
				/* state = 1 --> ON */
			
				/* If the EXTRA button is pressed for more than 100 ms, play a HIT sound */
				if (GetButtonTime(&button_extra, BUTTON_RELEASED, 100))
				{
					PlayRandomAlt(SND_TYPE_HIT);
				}
				
				if (accOnPulse())
				{
					/* Play a STRIKE sound on accelerometer pulse */
					PlayRandomAlt(SND_TYPE_STRIKE);
				} else if (accOnTransient())
				{
					/* Play a SWING sound on accelerometer transient */

					if (alt_playback.do_play && alt_playback.type == SND_TYPE_STRIKE)
					{
						/* Ignore SWING when doing a STRING sound */
					} else {
						PlayRandomAlt(SND_TYPE_SWING);
					}
				}		

				/* If the ON/OFF button is pressed for more that 500ms, turn the thing OFF */
				if (GetButtonTime(&button_on, BUTTON_PRESSED, 500))
				{
					/* Stop HUM */
					StopHum();
					
					/* Play the OFF sound */
					PlayOff();
					
					/* State = OFF */
					state = 0;
				}
				break;
		}
	}

LOAD:
	/* Stop the timers */
	sysStopTimer();
	
	/* Receive the configuration through the UART */
	configReceive();
	return 0;
}
