/***************************************************************************
 * Artekit STM32 DIY lightsaber audio board
 * Copyright (C) 2013 Artekit Italy
 * Written by Ivan Meleca
 * http://www.artekit.eu
 
### mma8452.c
 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#include "config.h"
#include "i2c.h"

/* Accelerometer I2C address */
#define MMA8452_ADDR 0x1C

/* Set to 1 when the accelerometer is ready */
static uint8_t accStarted = 0;

/* Set to 1 on pulse detect */
static volatile uint8_t accPulse = 0;

/* Set to 1 on transient detect */
static volatile uint8_t accTransient = 0;

/* Activates the accelerometer by writing the bit 1 on the 0x2A register */
static void accActivate(uint8_t activate)
{
	uint8_t reg = i2cReadRegister(MMA8452_ADDR, 0x2A);
	i2cWriteRegister(MMA8452_ADDR, 0x2A, activate ? reg | 1 : reg & ~1);
}

/* Set a 2g, 4g or 8g scale */
static void accSetScale(uint8_t scale)
{
	uint8_t reg;

	if (scale != 2 && scale != 4 && scale != 8) return;

	scale >>= 2;

	reg = i2cReadRegister(MMA8452_ADDR, 0x0E);
	reg &= ~0x03;
	reg |= scale;
	i2cWriteRegister(MMA8452_ADDR, 0x0E, reg);

	/* Set high-res */
	reg = i2cReadRegister(MMA8452_ADDR, 0x2B);
	reg |= 0x03;
	i2cWriteRegister(MMA8452_ADDR, 0x2B, reg);
}

/* Sets the output data rate */
static void accSetODR(uint16_t odr)
{
	uint8_t reg;

	if (odr != 800 && odr != 400 && odr != 200 && odr != 100 &&
		odr != 50 && odr != 12 && odr != 6 && odr != 1) return;

	reg = i2cReadRegister(MMA8452_ADDR, 0x2A);
	reg &= ~0x38;

	switch (odr)
	{
		case 800: break;
		case 400: reg |= (1 << 3); break;
		case 200: reg |= (2 << 3); break;
		case 100: reg |= (3 << 3); break;
		case 50: reg |= (4 << 3); break;
		case 12: reg |= (5 << 3); break;
		case 6: reg |= (6 << 3); break;
		case 1: reg |= (7 << 3); break;
	}

	i2cWriteRegister(MMA8452_ADDR, 0x2A, reg);
}

/* Sets the polarity of the interrupt pins */
static void accSetIntPolarity(uint8_t pol)
{
	uint8_t reg;
	
	if (pol > 1) return;

	reg = i2cReadRegister(MMA8452_ADDR, 0x2C);
	reg &= ~0x02;
	reg |= (pol << 1) & 0x02;
	i2cWriteRegister(MMA8452_ADDR, 0x2C, reg);
}

/* Activate both transient and pulse detection */
static void accActivateTransientAndPulse(uint8_t activate)
{
	uint8_t reg;
	
	reg = i2cReadRegister(MMA8452_ADDR, 0x2D);
	
	if (activate)
	{
		reg |= 0x28;
	} else {
		reg &= ~0x28;
	}
		
	i2cWriteRegister(MMA8452_ADDR, 0x2D, reg);
	
	if (activate)
	{
		/* Pins = Transient on INT1, Pulse on INT2 */
		reg = i2cReadRegister(MMA8452_ADDR, 0x2E);
		reg |= 0x20;
		i2cWriteRegister(MMA8452_ADDR, 0x2E, reg);
	}
}

/* Configures the 'single tap' pulse */
static void accConfigureSingleTap(void)
{
	/* The following values are being set with an ODR of 200KHz, with the LPF enabled
	 * and with oversampling (MODS bits in the 0x2B register).
	 * If you are looking at the tables on the reference manuals or app. notes, time
	 * steps are those of the "200" row, "HighRes" column.
	 */
	
	/* 0x0F: HP_FILTER_CUTOFF High-Pass Filter Register */
	/* Active low pass filter for pulse detection */
	i2cWriteRegister(MMA8452_ADDR, 0x0F, 0x10);
	
	/* 0x21: PULSE_CFG Pulse Configuration Register */
	/* Activate single tap on the three axes */
	i2cWriteRegister(MMA8452_ADDR, 0x21, 0x15);
	
	/* 0x23 - 0x25: PULSE_THSX, Y, Z Pulse Threshold for X, Y & Z Registers */
	/* 64 counts of acceleration on any axis + timing (see below) will trigger a 'pulse' */
	i2cWriteRegister(MMA8452_ADDR, 0x23, 0x40);
	i2cWriteRegister(MMA8452_ADDR, 0x24, 0x40);
	i2cWriteRegister(MMA8452_ADDR, 0x25, 0x40);
	
	/* Register 0x26: PULSE_TMLT Pulse Time Window 1 Register */
	/* 200ms of a 2g acceleration will trigger a 'pulse' */
	i2cWriteRegister(MMA8452_ADDR, 0x26, 0xC8);
	
	/* Register 0x27: PULSE_LTCY Pulse Latency Timer Register */
	/* Set 20ms as the time between two 'pulses' */
	i2cWriteRegister(MMA8452_ADDR, 0x27, 0x14);
}

/* Configures the transient detection values */
static void accConfigureTransient(void)
{
	/* Register 0x1D:TRANSIENT_CFG Transient Configuration Register */
	/* Activate detection on three axis, HPF enabled, LATCH disabled */
	i2cWriteRegister(MMA8452_ADDR, 0x1D, 0x0E);
	
	/* Register 0x1F: TRANSIENT_THS Transient Threshold Register */
	/* 0,35g acceleration triggers a transient interrupt (0,35g / 0,063) */
	i2cWriteRegister(MMA8452_ADDR, 0x1F, 5);
	
	/* Register 0x20: TRANSIENT_COUNT Transient Debounce Register */
	/* 50ms with on an 1g acceleration triggers an interrupt (50ms / 2.5) */
	i2cWriteRegister(MMA8452_ADDR, 0x20, 20);
}

/* Initial accelerometer configuration */
static void accConfigure(void)
{
	uint8_t reg;
	
	if (accStarted) return;

	/* Check for the right IC */
	reg = i2cReadRegister(MMA8452_ADDR, 0x0D);
	if (reg != 0x2A)
		return;

	/* De-activate */
	accActivate(0);
	
	/* All the values set here and in the accConfigureSingleTap() and accConfigureTransient()
	 * may be stored in the serial flash to personalize the sensitivity of both pulses
	 * and transients.
	 */
	
	/* 2g scale */
	accSetScale(2);
	
	/* 200 Hz output data rate */
	accSetODR(200);
	
	/* Interrupt polarity high */
	accSetIntPolarity(1);
	
	accConfigureSingleTap();
	
	accConfigureTransient();
	
	/* Transient and pulse interrupts */
	accActivateTransientAndPulse(1);
	
	accActivate(1);
	
	accStarted = 1;
}

/* Call this function to initialize the accelerometer */
void accInit(void)
{
	GPIO_InitTypeDef GPIO_InitStructure;
	EXTI_InitTypeDef EXTI_InitStructure;
	NVIC_InitTypeDef NVIC_InitStructure;
	
	i2cInit(400000);
		
	/* Configure external interrupts on pins 1 and 2 of port B */
	GPIO_InitStructure.GPIO_Pin =  GPIO_Pin_1 | GPIO_Pin_2;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN_FLOATING;
	GPIO_Init(GPIOB, &GPIO_InitStructure);
	
	accConfigure();
	
	/* Configure EXTI Line1 to generate an interrupt on rising edge */
  	EXTI_InitStructure.EXTI_Line = EXTI_Line1;
  	EXTI_InitStructure.EXTI_Mode = EXTI_Mode_Interrupt;
  	EXTI_InitStructure.EXTI_Trigger = EXTI_Trigger_Rising;
  	EXTI_InitStructure.EXTI_LineCmd = ENABLE;
  	EXTI_Init(&EXTI_InitStructure);
	
	/* Enable the EXTI1 Interrupt */
  	NVIC_InitStructure.NVIC_IRQChannel = EXTI1_IRQn;
  	NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 5;
  	NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
  	NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  	NVIC_Init(&NVIC_InitStructure);
	
	/* Connect EXTI Line1 to PB1 */
  	GPIO_EXTILineConfig(GPIO_PortSourceGPIOB, GPIO_PinSource1);
	
	/* Configure EXTI Line2 to generate an interrupt on rising edge */
  	EXTI_InitStructure.EXTI_Line = EXTI_Line2;
  	EXTI_InitStructure.EXTI_Mode = EXTI_Mode_Interrupt;
  	EXTI_InitStructure.EXTI_Trigger = EXTI_Trigger_Rising;
  	EXTI_InitStructure.EXTI_LineCmd = ENABLE;
  	EXTI_Init(&EXTI_InitStructure);
	
	/* Enable the EXTI2 Interrupt */
  	NVIC_InitStructure.NVIC_IRQChannel = EXTI2_IRQn;
  	NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 5;
  	NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
  	NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  	NVIC_Init(&NVIC_InitStructure);
	
	/* Connect EXTI Line1 to PB1 */
  	GPIO_EXTILineConfig(GPIO_PortSourceGPIOB, GPIO_PinSource2);
}

/* Checks if there was a pulse */
uint8_t accOnPulse(void)
{
	if (accPulse)
	{
		accPulse = 0;
		return 1;
	}
	return 0;
}

/* Checks if there was a transient */
uint8_t accOnTransient(void)
{
	if (accTransient)
	{
		accTransient = 0;
		return 1;
	}
	return 0;
}

/* Reset the pulse and transient status variables */
void accResetStatusFlags(void)
{
	accPulse = 0;
	accTransient = 0;
}

/* EXT1 interrupt handler (INT1 for transients) */
void EXTI1_IRQHandler(void)
{
	EXTI_ClearFlag(EXTI_Line1);
	accTransient = 1;	
}

/* EXT2 interrupt handler (INT2 for pulses) */
void EXTI2_IRQHandler(void)
{
	EXTI_ClearFlag(EXTI_Line2);	
	accPulse = 1;
}
